package de.ozgcloud.apilib.common.command.grpc;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.Mockito;

import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextProvider;
import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandStatus;
import de.ozgcloud.apilib.common.command.OzgCloudCreateSubCommandsRequestTestFactory;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.grpc.command.CommandServiceGrpc.CommandServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.command.GrpcAddSubCommandsRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcCommand;
import de.ozgcloud.vorgang.grpc.command.GrpcCommandsResponse;

class GrpcOzgCloudCommandServiceTest {

	private GrpcOzgCloudCommandService service;

	@Mock
	private CommandServiceBlockingStub serviceStub;
	@Mock
	private CommandMapper mapper;
	@Mock
	private OzgCloudCallContextProvider contextProvider;

	@BeforeEach
	void init() {
		service = spy(new GrpcOzgCloudCommandService(serviceStub, mapper, contextProvider, 600));
	}

	@Nested
	class TestCreateAndWaitUntilDone {
		private OzgCloudCommand command = OzgCloudCommandTestFactory.create();

		@Test
		void shouldCallCreate() {
			var created = OzgCloudCommandTestFactory.create();
			doReturn(created).when(service).create(any());

			service.createAndWaitUntilDone(command);

			verify(service).create(command);
		}

		@Test
		void shouldWaitUntilDone() {
			var created = OzgCloudCommandTestFactory.create();
			doReturn(created).when(service).create(any());

			service.createAndWaitUntilDone(command);

			verify(service).waitUntilDone(created);
		}

		@Test
		void shouldTerminateByTimeout() {
			var command = OzgCloudCommandTestFactory.createBuilder().status(OzgCloudCommandStatus.PENDING).build();
			doReturn(command).when(service).reloadCommand(any());

			service.waitUntilDone(command);

			Mockito.verify(service, Mockito.times(2)).reloadCommand(command.getId());
		}

		@Test
		void shouldCallVerifyCommand() {
			service.waitUntilDone(command);

			verify(service).verifyCommand(command);
		}

		@Nested
		class Create {

			@BeforeEach
			void init() {
				when(serviceStub.withInterceptors(any())).thenReturn(serviceStub);
				when(serviceStub.createCommand(any())).thenReturn(GrpcCommandResponseTestFactory.create());
			}

			@Test
			void shouldBuildRequest() {
				service.create(command);

				verify(mapper).buildCreateCommandRequest(command);
			}

			@Test
			void shouldCallStub() {
				var request = GrpcCreateCommandRequestTestFactory.create();
				doReturn(request).when(mapper).buildCreateCommandRequest(any());

				service.create(OzgCloudCommandTestFactory.create());

				verify(serviceStub).createCommand(request);
			}

			@Test
			void shouldMapCommand() {
				service.create(command);

				verify(mapper).fromGrpc(GrpcCommandTestFactory.create());
			}

			@Test
			void shouldReturnResult() {
				var expected = OzgCloudCommandTestFactory.create();
				when(mapper.fromGrpc(any())).thenReturn(expected);

				var returned = service.create(command);

				assertThat(returned).isSameAs(expected);
			}
		}
	}

	@Nested
	class TestStatus {

		@Test
		void shouldThrowException() {
			var command = OzgCloudCommandTestFactory.createBuilder().status(OzgCloudCommandStatus.from("ERROR")).build();

			assertThrows(TechnicalException.class, () -> service.verifyCommand(command));
		}

	}

	@Nested
	class TestAddSubCommands {

		private static final GrpcCommand GRPC_COMMAND = GrpcCommandTestFactory.create();

		@Mock
		private GrpcCommandsResponse commandsResponse;
		@Mock
		private GrpcAddSubCommandsRequest grpcAddSubCommandsRequest;

		@BeforeEach
		void init() {
			doReturn(serviceStub).when(service).getCommandServiceStub();
			when(serviceStub.addSubCommands(any())).thenReturn(commandsResponse);
			when(commandsResponse.getCommandList()).thenReturn(List.of(GRPC_COMMAND));
		}

		@Test
		void shouldCallGetCommandServiceStub() {
			addSubCommands();

			verify(service).getCommandServiceStub();
		}

		@Test
		void shouldCallMapToAddSubcommandsRequest() {
			var request = OzgCloudCreateSubCommandsRequestTestFactory.create();

			service.addSubCommands(request);

			verify(mapper).toAddSubCommandsRequest(request);
		}

		@Test
		void shouldCallAddSubCommands() {
			when(mapper.toAddSubCommandsRequest(any())).thenReturn(grpcAddSubCommandsRequest);

			addSubCommands();

			verify(serviceStub).addSubCommands(grpcAddSubCommandsRequest);
		}

		@Test
		void shouldCallMapFromGrpc() {
			addSubCommands();

			verify(mapper).fromGrpc(GRPC_COMMAND);
		}

		@Test
		void shouldReturnResult() {
			var createdCommand = OzgCloudCommandTestFactory.create();
			when(mapper.fromGrpc(any())).thenReturn(createdCommand);

			var result = addSubCommands();

			assertThat(result).containsExactly(createdCommand);
		}

		private List<OzgCloudCommand> addSubCommands() {
			return service.addSubCommands(OzgCloudCreateSubCommandsRequestTestFactory.create());
		}
	}
}